using System;
using System.IO;
using System.Collections;
using System.Collections.Generic;
using CookComputing.XmlRpc;

public struct XenAPIResponseStruct
{
  public string Status;
  [XmlRpcMissingMapping(MappingAction.Ignore)]
  public string Value;
  [XmlRpcMissingMapping(MappingAction.Ignore)]
  public string[] ErrorDescription;
}

public struct XenAPIResponseStructMV
{
  public string Status;
  [XmlRpcMissingMapping(MappingAction.Ignore)]
  public string[] Value;
  [XmlRpcMissingMapping(MappingAction.Ignore)]
  public string[] ErrorDescription;
}

public struct XenAPIResponseStructRecord
{
    public string Status;
    [XmlRpcMissingMapping(MappingAction.Ignore)]
    public XmlRpcStruct Value;
    [XmlRpcMissingMapping(MappingAction.Ignore)]
    public string[] ErrorDescription;
}

public struct XenAPIRequestXmMigrateOptions
{
    public int port;
    public int resource;
}

public interface IXenAPI : IXmlRpcProxy
{
    [XmlRpcMethod("session.login_with_password")]
    XenAPIResponseStruct session_login_with_password(string username, string password);

    [XmlRpcMethod("session.logout")]
    void session_logout(string session_id);

    [XmlRpcMethod("session.get_this_host")]
    XenAPIResponseStruct session_get_this_host(string session_id, string session_id2);

    [XmlRpcMethod("host.get_resident_VMs")]
    XenAPIResponseStructMV host_get_resident_VMs(string session_id, string host);

    [XmlRpcMethod("host.dmesg")]
    XenAPIResponseStruct host_dmesg(string session_id, string host);

    [XmlRpcMethod("host.dmesg_clear")]
    XenAPIResponseStruct host_dmesg_clear(string session_id, string host);

    [XmlRpcMethod("host.get_log")]
    XenAPIResponseStruct host_get_log(string session_id, string host);

    [XmlRpcMethod("host.list_methods")]
    XenAPIResponseStructMV host_list_methods(string session_id);

    [XmlRpcMethod("host.get_uuid")]
    XenAPIResponseStruct host_get_uuid(string session_id, string host);

    [XmlRpcMethod("host.get_name_label")]
    XenAPIResponseStruct host_get_name_label(string session_id, string host);

    [XmlRpcMethod("host.get_name_description")]
    XenAPIResponseStruct host_get_name_description(string session_id, string host);

    [XmlRpcMethod("VM.clone")]
    XenAPIResponseStruct VM_clone(string session_id, string VM_ref, string new_name);

    [XmlRpcMethod("VM.start")]
    XenAPIResponseStruct VM_start(string session_id, string VM_ref, bool start_paused);

    [XmlRpcMethod("VM.pause")]
    XenAPIResponseStruct VM_pause(string session_id, string VM_ref);

    [XmlRpcMethod("VM.unpause")]
    XenAPIResponseStruct VM_unpause(string session_id, string VM_ref);

    [XmlRpcMethod("VM.clean_shutdown")]
    XenAPIResponseStruct VM_clean_shutdown(string session_id, string VM_ref);

    [XmlRpcMethod("VM.clean_reboot")]
    XenAPIResponseStruct VM_clean_reboot(string session_id, string VM_ref);

    [XmlRpcMethod("VM.hard_shutdown")]
    XenAPIResponseStruct VM_hard_shutdown(string session_id, string VM_ref);

    [XmlRpcMethod("VM.hard_reboot")]
    XenAPIResponseStruct VM_hard_reboot(string session_id, string VM_ref);

    [XmlRpcMethod("VM.suspend")]
    XenAPIResponseStruct VM_suspend(string session_id, string VM_ref);

    [XmlRpcMethod("VM.resume")]
    XenAPIResponseStruct VM_resume(string session_id, string VM_ref, bool start_paused);

    [XmlRpcMethod("VM.migrate")]
    XenAPIResponseStruct VM_migrate(string session_id, string VM_ref, string dest, bool live, XenAPIRequestXmMigrateOptions options);

    [XmlRpcMethod("VM.get_record")]
    XenAPIResponseStructRecord VM_get_record(string session_id, string VM_ref);

    [XmlRpcMethod("VM.get_domid")]
    XenAPIResponseStruct VM_get_domid(string session_id, string VM_ref);

    [XmlRpcMethod("VM.get_is_control_domain")]
    XenAPIResponseStruct VM_get_is_control_domain(string session_id, string VM_ref);

    [XmlRpcMethod("VM.get_metrics")]
    XenAPIResponseStruct VM_get_metrics(string session_id, string VM_ref);

    [XmlRpcMethod("VM.get_power_state")]
    XenAPIResponseStruct VM_get_power_state(string session_id, string VM_ref);

    [XmlRpcMethod("VM.get_name_label")]
    XenAPIResponseStruct VM_get_name_label(string session_id, string VM_ref);

    [XmlRpcMethod("VM.get_by_name_label")]
    XenAPIResponseStruct get_by_name_label(string session_id);

    [XmlRpcMethod("VM.get_name_description")]
    XenAPIResponseStruct VM_get_name_description(string session_id, string VM_ref);

    [XmlRpcMethod("VM.get_memory_static_max")]
    XenAPIResponseStruct VM_get_memory_static_max(string session_id, string VM_ref);

    [XmlRpcMethod("VM.get_memory_dynamic_max")]
    XenAPIResponseStruct VM_get_memory_dynamic_max(string session_id, string VM_ref);

    [XmlRpcMethod("VM.get_memory_dynamic_min")]
    XenAPIResponseStruct VM_get_memory_dynamic_min(string session_id, string VM_ref);

    [XmlRpcMethod("VM.get_memory_static_min")]
    XenAPIResponseStruct VM_get_memory_static_min(string session_id, string VM_ref);

    [XmlRpcMethod("VM.get_VCPUs_max")]
    XenAPIResponseStruct VM_get_VCPUs_max(string session_id, string VM_ref);

    [XmlRpcMethod("VM.get_VCPUs_at_startup")]
    XenAPIResponseStruct VM_get_VCPUs_at_startup(string session_id, string VM_ref);

    [XmlRpcMethod("VM_metrics.get_record")]
    XenAPIResponseStructRecord VM_metrics_get_record(string session_id, string VM_metrics_ref);

    [XmlRpcMethod("VM_metrics.get_uuid")]
    XenAPIResponseStruct VM_metrics_get_uuid(string session_id, string VM_metrics_ref);

    [XmlRpcMethod("VM_metrics.get_memory_actual")]
    XenAPIResponseStruct VM_metrics_get_memory_actual(string session_id, string VM_metrics_ref);

    [XmlRpcMethod("VM_metrics.get_VCPUs_number")]
    XenAPIResponseStruct VM_metrics_get_VCPUs_number(string session_id, string VM_metrics_ref);

    [XmlRpcMethod("VM_metrics.get_VCPUs_utilisation")]
    XenAPIResponseStruct VM_metrics_get_VCPUs_utilisation(string session_id, string VM_metrics_ref);

    [XmlRpcMethod("VM_metrics.get_VCPUs_CPU")]
    XenAPIResponseStruct VM_metrics_get_VCPUs_CPU(string session_id, string VM_metrics_ref);

    [XmlRpcMethod("VM_metrics.get_state")]
    XenAPIResponseStructMV VM_metrics_get_state(string session_id, string VM_metrics_ref);
    
}

class XenAPI
{
    private IXenAPI xend;
    private string session;
    private string username;
    private string password;
    public int debug;
    private ArrayList VMs = new ArrayList();

    public XenAPI()
    {
        try
        {
            this.xend = XmlRpcProxyGen.Create<IXenAPI>();

			/* for debug
            RequestResponseLogger dumper = new RequestResponseLogger();
            dumper.Directory = "C:/temp";
            dumper.Attach(xend);
            */
        }
        catch (XmlRpcFaultException fex)
        {
            Console.WriteLine(fex.FaultString);
        }
        catch (Exception ex)
        {
            Console.WriteLine(ex.Message);
        }
    }

    public bool SetUrl(string url)
    {
        this.xend.Url = url;
        return true;
    }

    public bool SetUrl(string protocol, string host, int port)
    {
        this.xend.Url = String.Format("{0}://{1}:{2}/", protocol, host, port.ToString());
        return true;
    }

    public string GetUrl()
    {
        return this.xend.Url;
    }

    public string Login(string username, string password)
    {
        XenAPIResponseStruct res;

        try
        {
            res = this.xend.session_login_with_password(username, password);
            if (res.Status == "Failure" && res.ErrorDescription[0] == "SESSION_AUTHENTICATION_FAILED")
            {
                Console.WriteLine("ERROR: Login failure. Please check your password.");
            }
            else if (res.Status != "Success")
            {
                Console.WriteLine("ERROR: Login failure.");
                return null;
            }

            this.session = res.Value;
            this.username = username;
            this.password = password;
        }
        catch (System.Net.WebException wex)
        {
            Console.WriteLine("ERROR: Cann't connect to xend.");
            return null;
        }
        catch (Exception ex)
        {
            Console.WriteLine("ERROR: Cann't connect to xend.");
            return null;
        }

        return this.session;
    }

    public string ResumeSession(string url, string session)
    {
        this.SetUrl(url);
        this.session = session;

        if (this.GetHost() == null) return null;

        return this.session;
    }

    public bool Logout()
    {
        this.xend.session_logout(this.session);
        return true;
    }

    public string Relogin() {
        string new_session;

        this.Logout();

        new_session = this.Login(this.username, this.password);
        if (new_session == null)
        {
            Console.WriteLine("ERROR: Cann't reconnect to xend.");
            return null;
        }
        return new_session;
    }

    public string GetHost()
    {
        XenAPIResponseStruct res;

        res = this.xend.session_get_this_host(this.session, this.session);

        if (res.Status == "Failure" && res.ErrorDescription[0] == "SESSION_INVALID")
        {
            return null;
        }
        else if (res.Status != "Success")
        {

            return null;
        }
        return res.Value;
    }

    public bool ScanVM()
    {
        Host host = new Host(this);
        string[] vmlist = host.GetResidentVMs();
        if (vmlist == null) return false;

        this.VMs.Clear();
        foreach (string v in vmlist)
        {
            this.VMs.Add(new VM(this, v));
        }

        return true;
    }

    public ArrayList GetAllVMs()
    {
        return this.VMs;
    }

    public class Host
    {
        XenAPI xenapi;
        
        public Host(XenAPI xenapi)
        {
            this.xenapi = xenapi;
        }

        public string[] GetResidentVMs()
        {
            XenAPIResponseStructMV res;
            String host;
            host = this.xenapi.GetHost();
            if (host == null) return null;

            res = this.xenapi.xend.host_get_resident_VMs(this.xenapi.session, host);

            if (res.Status != "Success")
            {
                return null;
            }
            return res.Value;
        }

        public string GetDmesg(bool clear_buffer)
        {
            XenAPIResponseStruct res;
            String host;
            host = this.xenapi.GetHost();
            if (host == null) return null;

            if (clear_buffer)
            {
                res = this.xenapi.xend.host_dmesg_clear(this.xenapi.session, host);
            }
            else
            {
                res = this.xenapi.xend.host_dmesg(this.xenapi.session, host);
            }

            if (res.Status != "Success")
            {
                    return null;
            }

            return res.Value;
        }

        public string GetLog()
        {
            XenAPIResponseStruct res;
            String host;
            host = this.xenapi.GetHost();
            if (host == null) return null;

            res = this.xenapi.xend.host_get_log(this.xenapi.session, host);
            if (res.Status != "Success")
            {
                return null;
            }
            return res.Value;
        }

    }

    public class VM
    {
        XenAPI xenapi;
        private string _ref;

        public string uuid;
        public string powerState;
        public string nameLabel;
        public string nameDescription;
        public long userVersion = -1;
        public bool isATemplate;
        public bool autoPowerOn;
        // public VDI suspendVDI;
        // public Host residentOn;
        public long memoryStaticMax;
        public long memoryDynamicMax;
        public long memoryStaticMin;
        public long memoryDynamicMin;
        // public Map<String, String> VCPUsParams;
        public Dictionary<string, string> VCPUsParams = new Dictionary<string, string>();
        public long VCPUsMax;
        public long VCPUsAtStartup;
        // public Types.OnNormalExit actionsAfterShutdown;
        // public Types.OnNormalExit actionsAfterReboot;
        // public Types.OnCrashBehaviour actionsAfterCrash;
        // public Set<Console> consoles;
        // public Set<VIF> VIFs;
        // public Set VDB VDBs;
        // public Set<CrashDump> crashDump;
        // public Set<VTPM> VTPMs;
        public string PVKernel;
        public string PVRamdisk;
        public string PVArgs;
        public string PVBootloaderArgs;
        public string HVMBootPolicy;
        //public Map<String, String> HVMBootParams;
        public Dictionary<string, string> HVMBootParams;
        //public Map<String, String> platform;
        public Dictionary<string, string> platform;
        public string PCIBus;
        //public Map<String, String> otherConfig;
        public Dictionary<string, string> otherConfig;
        public long domid;
        public bool isControlDomain;
        public VMMetrics metrics; // ȑO
        //public VMGuestMetrics guestMetrics;


        public VM(XenAPI xenapi, string _ref)
        {
            this.xenapi = xenapi;
            this._ref = _ref;
            this.uuid = _ref;
            this.Sync();
            this.metrics = this.GetVMMetrics();
        }

        public string GetUUID()
        {
            return this.uuid;
        }

        public void Sync()
        {
            XenAPIResponseStructRecord res;

            res = this.xenapi.xend.VM_get_record(this.xenapi.session, this._ref);
            if (res.Status != "Success")
            {
                Console.WriteLine("ERR: Sync();");
                return;
            }

            foreach (string key in res.Value.Keys)
            {
                switch (key)
                {
                    case "uuid":
                        this.uuid = res.Value[key].ToString();
                        break;
                    case "power_state":
                        this.powerState = res.Value[key].ToString();
                        break;
                    case "name_label":
                        this.nameLabel = res.Value[key].ToString();
                        break;
                    case "name_description":
                        this.nameDescription = res.Value[key].ToString();
                        break;
                    case "user_version":
                        this.userVersion = long.Parse(res.Value[key].ToString());
                        break;
                    case "is_a_template":
                        this.isATemplate = bool.Parse(res.Value[key].ToString());
                        break;
                    case "auto_power_on":
                        this.autoPowerOn = bool.Parse(res.Value[key].ToString());
                        break;
//                    case "resident_on":
//                        this.residentOn = long.Parse(res.Value[key].ToString());
//                        break;
                    case "memory_static_max":
                        this.memoryStaticMax = long.Parse(res.Value[key].ToString());
                        break;
                    case "memory_dynamic_max":
                        this.memoryDynamicMax = long.Parse(res.Value[key].ToString());
                        break;
                    case "memory_static_min":
                        this.memoryStaticMin = long.Parse(res.Value[key].ToString());
                        break;
                    case "memory_dynamic_min":
                        this.memoryDynamicMin = long.Parse(res.Value[key].ToString());
                        break;
                    case "VCPUs_params":
 //                               public Dictionary<string, string> VCPUsParams = new Dictionary<string, string>();
                        break;
                    case "VCPUs_max":
                        this.VCPUsMax = long.Parse(res.Value[key].ToString());
                        break;
                    case "VCPUs_at_startup":
                        this.VCPUsAtStartup = long.Parse(res.Value[key].ToString());
                        break;
                    case "actions_after_reboot":
                    case "actions_after_suspend":
                    case "actions_after_shutdown":
                    case "actions_after_crash":
// public Types.OnNormalExit actionsAfterShutdown;
// public Types.OnNormalExit actionsAfterReboot;
// public Types.OnCrashBehaviour actionsAfterCrash;
                        break;
                    case "consoles":
                    case "VIFs":
                    case "VDBs":
                    case "crashDump":
                    case "VTPMs":
// public Set<Console> consoles;
// public Set<VIF> VIFs;
// public Set<VDB> VDBs;
// public Set<CrashDump> crashDump;
// public Set<VTPM> VTPMs;
                        break;
                    case "PV_kernel":
                        this.PVKernel = res.Value[key].ToString();
                        break;
                    case "PV_ramdisk":
                        this.PVRamdisk = res.Value[key].ToString();
                        break;
                    case "PV_args":
                        this.PVArgs = res.Value[key].ToString();
                        break;
                    case "PV_bootloader_args":
                        this.PVBootloaderArgs = res.Value[key].ToString();
                        break;
                    case "HVM_boot_policy":
                        this.HVMBootPolicy = res.Value[key].ToString();
                        break;
//                                public Dictionary<string, string> HVMBootParams = new Dictionary<string, string>();
//                                public Dictionary<string, string> platform = new Dictionary<string, string>();
                    case "PCI_bus":
                        this.PCIBus = res.Value[key].ToString();
                        break;
//                                public Dictionary<string, string> otherConfig = new Dictionary<string, string>();
                    case "domid":
                        this.domid = long.Parse(res.Value[key].ToString());
                        break;
                    case "is_control_domain":
                        this.isControlDomain = bool.Parse(res.Value[key].ToString());
                        break;
                }
            }
        }

        public void GetRecord()
        {
            XenAPIResponseStructRecord res;

            res = this.xenapi.xend.VM_get_record(this.xenapi.session, this._ref);
            if (res.Status != "Success")
            {
                Console.WriteLine("ERR: ShowRecord();");
                return;
            }

            foreach (DictionaryEntry d in res.Value)
            {
                Console.WriteLine("Key:{0}, Value:{1}", d.Key, d.Value);
            }
        }

        public long GetDomID()
        {
            XenAPIResponseStruct res;
            res = this.xenapi.xend.VM_get_domid(this.xenapi.session, this._ref);

            if (res.Status == "Success")
            {
                this.domid = long.Parse(res.Value);
            } else {
                return -1;
            }

            return this.domid;
        }

        public void ShowRecord()
        {
            XenAPIResponseStructRecord res;

            res = this.xenapi.xend.VM_get_record(this.xenapi.session, this._ref);
            if (res.Status != "Success")
            {
                Console.WriteLine("ERR: ShowRecord();");
                return;
            }

            foreach (DictionaryEntry d in res.Value)
            {
                Console.WriteLine("Key:{0}, Value:{1}", d.Key, d.Value);
            }
        }

        public bool CleanShutdown()
        {
            XenAPIResponseStruct res;
            res = this.xenapi.xend.VM_clean_shutdown(this.xenapi.session, this._ref);

            if (res.Status != "Success")
            {
                return false;
            }

            return true;
        }

        public bool HardShutdown()
        {
            XenAPIResponseStruct res;
            res = this.xenapi.xend.VM_hard_shutdown(this.xenapi.session, this._ref);

            if (res.Status != "Success")
            {
                return false;
            }

            return true;
        }

        public bool Start()
        {
            return this.Start(false);
        }

        public bool Start(bool start_paused)
        {
            XenAPIResponseStruct res;
            res = this.xenapi.xend.VM_start(this.xenapi.session, this._ref, start_paused);

            if (res.Status != "Success")
            {
                return false;
            }

            return true;
        }

        public bool CleanReboot()
        {
            XenAPIResponseStruct res;
            res = this.xenapi.xend.VM_clean_reboot(this.xenapi.session, this._ref);

            if (res.Status != "Success")
            {
                return false;
            }

            return true;
        }

        public bool HardReboot()
        {
            XenAPIResponseStruct res;
            res = this.xenapi.xend.VM_hard_reboot(this.xenapi.session, this._ref);

            if (res.Status != "Success")
            {
                return false;
            }

            return true;
        }
        public bool Suspend()
        {
            XenAPIResponseStruct res;
            res = this.xenapi.xend.VM_suspend(this.xenapi.session, this._ref);

            if (res.Status != "Success")
            {
                return false;
            }

            return true;
        }

        public bool Resume()
        {
            return this.Resume(false);
        }

        public bool Resume(bool start_paused)
        {
            XenAPIResponseStruct res;
            res = this.xenapi.xend.VM_resume(this.xenapi.session, this._ref, start_paused);

            if (res.Status != "Success")
            {
                return false;
            }

            return true;
        }

        public bool Pause()
        {
            XenAPIResponseStruct res;
            res = this.xenapi.xend.VM_pause(this.xenapi.session, this._ref);

            if (res.Status != "Success")
            {
                return false;
            }

            return true;
        }

        public bool Unpause()
        {
            XenAPIResponseStruct res;
            res = this.xenapi.xend.VM_unpause(this.xenapi.session, this._ref);

            if (res.Status != "Success")
            {
                return false;
            }

            return true;
        }

        public bool Migrate(string dest, int port, bool live)
        {
            XenAPIRequestXmMigrateOptions options;
            options.resource = 0;
            options.port = port;

            XenAPIResponseStruct res;
            res = this.xenapi.xend.VM_migrate(this.xenapi.session, this._ref, dest, live, options);

            if (res.Status == "Failure" && res.ErrorDescription[0] == "VM_BAD_POWER_STATE")
            {
                Console.WriteLine("ERR: The domain is bad power state.");
                return false;
            }
            else if (res.Status == "Failure" && res.ErrorDescription[0] == "INTERNAL_ERROR")
            {
                Console.WriteLine("ERR: {0}", res.ErrorDescription[1]);
                return false;
            }
            else if(res.Status != "Success")
            {
                return false;
            }

            return true;
        }

        public string GetNameLabel()
        {
            XenAPIResponseStruct res;
            res = this.xenapi.xend.VM_get_name_label(this.xenapi.session, this._ref);

            if (res.Status == "Success")
            {
                this.nameLabel = res.Value;
            }
            else
            {
                return null;
            }

            return this.nameLabel;
        }

        public string GetNameDescription()
        {
            XenAPIResponseStruct res;
            res = this.xenapi.xend.VM_get_name_description(this.xenapi.session, this._ref);


            if (res.Status == "Success")
            {
                this.nameDescription = res.Value;
            }
            else
            {
                return null;
            }

            return this.nameDescription;
        }

        public string GetPowerState()
        {
            XenAPIResponseStruct res;
            res = this.xenapi.xend.VM_get_power_state(this.xenapi.session, this._ref);

            if (res.Status == "Success")
            {
                this.powerState = res.Value;
            }
            else
            {
                return null;
            }

            return this.powerState;
        }

        public long GetMemoryStaticMax()
        {
            XenAPIResponseStruct res;
            res = this.xenapi.xend.VM_get_memory_static_max(this.xenapi.session, this._ref);

            if (res.Status != "Success")
            {
                return -1;
            }
            return long.Parse(res.Value);
        }

        public long GetMemoryStaticMin()
        {
            XenAPIResponseStruct res;
            res = this.xenapi.xend.VM_get_memory_static_min(this.xenapi.session, this._ref);

            if (res.Status != "Success")
            {
                return -1;
            }
            return long.Parse(res.Value);
        }

        public long GetMemoryDynamicMax()
        {
            XenAPIResponseStruct res;
            res = this.xenapi.xend.VM_get_memory_dynamic_max(this.xenapi.session, this._ref);

            if (res.Status != "Success")
            {
                return -1;
            }
            return long.Parse(res.Value);
        }

        public long GetMemoryDynamicMin()
        {
            XenAPIResponseStruct res;
            res = this.xenapi.xend.VM_get_memory_dynamic_min(this.xenapi.session, this._ref);

            if (res.Status != "Success")
            {
                return -1;
            }
            return long.Parse(res.Value);
        }

        public long GetVCPUsMax()
        {
            XenAPIResponseStruct res;
            res = this.xenapi.xend.VM_get_VCPUs_max(this.xenapi.session, this._ref);

            if (res.Status != "Success")
            {
                return -1;
            }
            return long.Parse(res.Value);
        }

        public long GetVCPUsAtStartup()
        {
            XenAPIResponseStruct res;
            res = this.xenapi.xend.VM_get_VCPUs_at_startup(this.xenapi.session, this._ref);

            if (res.Status == "Success")
            {
                this.VCPUsAtStartup = long.Parse(res.Value);
            }
            else if (res.Status == "Failure")
            {
                this.VCPUsAtStartup = -1;
            }

            return this.VCPUsAtStartup;
        }

        public bool GetIsControlDomain()
        {
            XenAPIResponseStruct res;
            res = this.xenapi.xend.VM_get_is_control_domain(this.xenapi.session, this._ref);

            if (res.Status == "Success")
            {
                this.isControlDomain = bool.Parse(res.Value);
            }

            return this.isControlDomain;
        }

        public VMMetrics GetVMMetrics()
        {
            if (this.metrics == null)
            {
                XenAPIResponseStruct res;
                res = this.xenapi.xend.VM_get_metrics(this.xenapi.session, this._ref);

                if (res.Status != "Success")
                {
                    return null;
                }

                return new VMMetrics(this.xenapi, res.Value);
            }

            return this.metrics;
        }
    }

    public class VMMetrics
    {
        XenAPI xenapi;
        private string _ref;
        public string uuid;
        public long memoryActual;
        public long VCPUsNumber;
        public Dictionary<long, double> VCPUsUtilisation;
        public Dictionary<long, long> VCPUsCPU;
        public Dictionary<string, string> VCPUsParams;
        public Dictionary<long, String[]> VCPUsFlags;
        public string[] state;
        public DateTime startTime;
        public DateTime lastUpdated;

        public VMMetrics(XenAPI xenapi, string _ref)
        {
            this._ref = _ref;
            this.xenapi = xenapi;
            this.uuid = _ref;
        }

        public void Sync()
        {
            XenAPIResponseStructRecord res;

            res = this.xenapi.xend.VM_metrics_get_record(this.xenapi.session, this._ref);
            if (res.Status != "Success")
            {
                Console.WriteLine("ERR: Sync();");
                return;
            }

            foreach (string key in res.Value.Keys)
            {
                switch (key)
                {
                    case "uuid":
                        this.uuid = res.Value[key].ToString();
                        break;
                    case "memory_actual":
                        this.memoryActual = long.Parse(res.Value[key].ToString());
                        break;
                    case "VCPUs_number":
                        this.VCPUsNumber = long.Parse(res.Value[key].ToString());
                        break;
                    case "VCPUs_utilization":
//                        this.VCPUsUtilization = ;
                        break;
                    case "VCPUs_CPU":
//                        this.VCPUsCPU = ;
                        break;
                    case "VCPUs_params":
                        //this.VCPUsParams = ;
                        break;
                    case "VCPUs_flags":
                        //this.VCPUsFlags = ;
                        break;
                    case "state":
                        //this.state = ;
                        break;
                    case "start_time":
                        //this.startTime = ;
                        break;
                    case "last_updated":
                        //this.lastUpdated = ;
                        break;
                }
            }
        }

        public string GetUUID()
        {
            return this.uuid;
        }

        public long GetMemoryActual()
        {
            XenAPIResponseStruct res;
            res = this.xenapi.xend.VM_metrics_get_memory_actual(this.xenapi.session, this.uuid);

            if (res.Status == "Success")
            {
                this.memoryActual = long.Parse(res.Value);
            }
            else
            {
                this.memoryActual = -1;
            }

            return this.memoryActual;
        }

        public long GetVCPUsNumber()
        {
            XenAPIResponseStruct res;
            res = this.xenapi.xend.VM_metrics_get_VCPUs_number(this.xenapi.session, this.uuid);

            if (res.Status == "Success")
            {
                this.VCPUsNumber = long.Parse(res.Value);
            }
            else
            {
                this.VCPUsNumber = -1;
            }

            return this.VCPUsNumber;
        }

        public string GetState()
        {
            XenAPIResponseStructMV res;
            res = this.xenapi.xend.VM_metrics_get_state(this.xenapi.session, this.uuid);

            if (res.Status == "Success" && res.Value.Length > 0)
            {
                return res.Value[0];
            } else {
                return null;
            }
        }

        public override string ToString()
        {
            return this._ref;
        }
    }
}