using System;
using System.IO;
using System.Collections;
using System.Text.RegularExpressions;

class xm
{
    public static int Main(string[] args)
    {
        XenAPI xenapi = new XenAPI();
        xenapi.debug = 0;

        if (args.Length == 1 && args[0] == "help")
        {
            help();
            return 1;
        }

        if (args.Length == 2 && args[0] == "connect")
        {
            string uri = args[1];
            if (connect(xenapi, uri) == null)
            {
                return 1;
            }
            else
            {
                return 0;
            }
        }
        else if (args.Length == 0)
        {
            help();
            return 1;
        }
        else
        {
            if (resume_session(xenapi) == null)
            {
                Console.WriteLine("ERROR: The session doesn't exist.");
                return 1;
            }
        }

        if (args.Length >= 1)
        {
            switch (args[0])
            {
                case "disconnect":
                    disconnect(xenapi);
                    break;
                case "domid":
                    if (args.Length == 2)
                    {
                        domid(xenapi, args[1]);
                    }
                    else
                    {
                        help();
                        return 1;
                    }
                    break;
                case "domname":
                    if (args.Length == 2)
                    {
                        domname(xenapi, args[1]);
                    }
                    else
                    {
                        help();
                        return 1;
                    }
                    break;
                case "list":
                    list(xenapi);
                    break;
                case "destroy":
                    if (args.Length == 2)
                    {
                        destroy(xenapi, args[1]);
                    }
                    else
                    {
                        help();
                        return 1;
                    }
                    break;
                case "shutdown":
                    if (args.Length == 2)
                    {
                        shutdown(xenapi, args[1]);
                    }
                    else
                    {
                        help();
                        return 1;
                    }
                    break;
                case "start":
                    if (args.Length == 2)
                    {
                        start(xenapi, args[1], false);
                    }
                    else if (args.Length == 3 &&
                        (args[1] == "-p" || args[1] == "--paused"))
                    {
                        start(xenapi, args[2], true);
                    }
                    else if (args.Length == 3 &&
                        (args[2] == "-p" || args[2] == "--paused"))
                    {
                        start(xenapi, args[1], true);
                    }
                    else
                    {
                        help();
                        return 1;
                    }
                    break;
                case "reboot":
                    if (args.Length == 2)
                    {
                        reboot(xenapi, args[1]);
                    }
                    else
                    {
                        help();
                        return 1;
                    }
                    break;
                case "suspend":
                    if (args.Length == 2)
                    {
                        suspend(xenapi, args[1]);
                    }
                    else
                    {
                        help();
                        return 1;
                    }
                    break;
                case "resume":
                    if (args.Length == 2)
                    {
                        resume(xenapi, args[1], false);
                    }
                    else if (args.Length == 3 &&
                        (args[1] == "-p" || args[1] == "--paused"))
                    {
                        resume(xenapi, args[2], true);
                    }
                    else if (args.Length == 3 &&
                        (args[2] == "-p" || args[2] == "--paused"))
                    {
                        resume(xenapi, args[1], true);
                    }
                    else
                    {
                        help();
                        return 1;
                    }
                    break;
                case "pause":
                    if (args.Length == 2)
                    {
                        pause(xenapi, args[1]);
                    }
                    else
                    {
                        help();
                        return 1;
                    }
                    break;
                case "unpause":
                    if (args.Length == 2)
                    {
                        unpause(xenapi, args[1]);
                    }
                    else
                    {
                        help();
                        return 1;
                    }
                    break;
                case "migrate":
                    int destPort = 8002;
                    if (args.Length == 3)
                    {
                        migrate(xenapi, args[1], args[2], destPort, false);
                    }
                    else if (args.Length == 4 &&
                        (args[1] == "-l" || args[1] == "--live"))
                    {
                        migrate(xenapi, args[2], args[3], destPort, true);
                    }
                    else
                    {
                        help();
                        return 1;
                    }
                    break;
                case "dmesg":
                    if (args.Length == 1)
                    {
                        dmesg(xenapi, false);
                    }
                    else if (args.Length == 2 &&
                        (args[1] == "-c" || args[1] == "--clear"))
                    {
                        dmesg(xenapi, true);
                    }
                    else
                    {
                        help();
                        return 1;
                    }
                    break;
                case "log":
                    log(xenapi);
                    break;
                default:
                    help();
                    return 1;
            }
        }

        return 0;
    }

    private static bool _IsDom0(string target)
    {
        if (target == "0" || target == "Domain-0")
        {
            return true;
        }
        else
        {
            return false;
        }
    }

    private static bool _IsTarget(XenAPI.VM vm, string target)
    {
        if (vm.domid.ToString() == target || vm.nameLabel == target)
        {
            return true;
        }
        else
        {
            return false;
        }
    }

    public static string _GenerateStateString(XenAPI.VM vm)
    {
        string result = "";

        if (vm.powerState == "Running")
            result = "r-----";

        return result;
    }

    public static string connect(XenAPI xenapi, string uri)
    {
        //        Regex RegexCheckUri = new Regex(@"^(https?)://(([\w\.\-]+)(:.+)?@)?(([a-zA-Z0-9\.\-]+)|((\d+)((\.(\d+)){3})))(:(\d+))?/?$");

        Regex RegexParseUri = new Regex(@"^(?<protocol>https?)://((?<username>[\w\.\-]+)(:(?<password>.+))?@)?(?<host>([a-zA-Z0-9\.\-]+)|((\d+)((\.(\d+)){3})))(:(?<port>(\d+)))?/?$");

        Match m = RegexParseUri.Match(uri);
        if(m == null) return null;

        if (m.Success)
        {
            string url;
            string protocol = "http";
            string username;
            string password;
            string host = "";
            string port = "9363";
            string session;

            protocol = m.Groups["protocol"].Value;
            username = m.Groups["username"].Value;
            password = m.Groups["password"].Value;
            host = m.Groups["host"].Value;

            if (m.Groups["port"].Value != "") port = m.Groups["port"].Value;
            if (m.Groups["port"].Value == "" && m.Groups["protocol"].Value == "https") port = "9394";

            url = String.Format("{0}://{1}:{2}/", protocol, host, port);
            xenapi.SetUrl(url);
            session = xenapi.Login(username, password);
            if (session != null)
            {
                Microsoft.Win32.RegistryKey regkey = Microsoft.Win32.Registry.CurrentUser.CreateSubKey(@"Volatile Environment");

                regkey.SetValue("XM_SESSION_URL", url);
                regkey.SetValue("XM_SESSION_ID", session);

                Console.WriteLine("XM_SESSION_URL: {0}", regkey.GetValue("XM_SESSION_URL"));
                Console.WriteLine("XM_SESSION_ID: {0}", regkey.GetValue("XM_SESSION_ID"));
                regkey.Close();
            }

            return session;
        }
        else
        {
            Console.WriteLine("ERROR: URL parse error.");
            return null;
        }
    }

    public static string resume_session(XenAPI xenapi)
    {
        string url;
        string session;

        Microsoft.Win32.RegistryKey regkey = Microsoft.Win32.Registry.CurrentUser.OpenSubKey(@"Volatile Environment");
        url = (string)regkey.GetValue("XM_SESSION_URL");
        session = (string)regkey.GetValue("XM_SESSION_ID");
        regkey.Close();
        if (url == null || session == null)
        {
            return null;
        }

        if (xenapi.ResumeSession(url, session) == null)
        {
            Console.WriteLine("ERROR: The session wasn't established. Please login again.");
            disconnect(xenapi);
            return null;
        }

        return session;
    }

    public static int disconnect(XenAPI xenapi)
    {
        xenapi.Logout();

        Microsoft.Win32.RegistryKey regkey = Microsoft.Win32.Registry.CurrentUser.CreateSubKey(@"Volatile Environment");
        regkey.DeleteValue("XM_SESSION_URL");
        regkey.DeleteValue("XM_SESSION_ID");
        regkey.Close();

        return 0;
    }

    /*
    public static void disconnect(XenAPI xenapi, string session)
    {
        Console.WriteLine("session: {0}", session);
        xenapi.Logout();
    }
    */

    public static int domid(XenAPI x, string target)
    {
        bool isSuccessScanVM;
        isSuccessScanVM = x.ScanVM();
        if (!isSuccessScanVM)
        {
            Console.WriteLine("ERROR: Could not get VMs information.");
            return 1;
        }

        ArrayList VMs;
        VMs = x.GetAllVMs();
        foreach (XenAPI.VM vm in VMs)
        {
            if(_IsTarget(vm, target))
            {
                if (vm.domid >= 0)
                {
                    Console.WriteLine(vm.domid);
                }
                else {
                    Console.WriteLine("None");
                }
                return 0;
            }
        }

        Console.WriteLine("ERROR: Domain '{0}' does not exist.", target);
        return 1;
    }

    public static int domname(XenAPI x, string target)
    {
        bool isSuccessScanVM = x.ScanVM();
        if (!isSuccessScanVM)
        {
            Console.WriteLine("ERROR: Could not get VMs information.");
            return 1;
        }

        ArrayList VMs;
        VMs = x.GetAllVMs();
        foreach (XenAPI.VM vm in VMs)
        {
            if (_IsTarget(vm, target))
            {
                Console.WriteLine(vm.nameLabel);
                return 0;
            }
        }

        Console.WriteLine("ERROR: Domain '{0}' does not exist.", target);
        return 1;
    }

    public static int list(XenAPI x)
    {
        bool isSuccessScanVM = x.ScanVM();
        if (!isSuccessScanVM)
        {
            Console.WriteLine("ERROR: Could not get VMs information.");
            return 1;
        }

        Console.Write("Name");
        for (int i = 0; i < 32 - 4; i++)
        {
            Console.Write(" ");
        }
        Console.WriteLine("ID\t Mem\tVCPUs\t State");

        ArrayList VMs;
        VMs = x.GetAllVMs();
        foreach (XenAPI.VM vm in VMs)
        {
            string domid = vm.domid.ToString();
            string state = _GenerateStateString(vm);

            if (domid == "-1") domid = "";

            Console.Write(vm.nameLabel);
            for (int i = 0; i < 32 - vm.nameLabel.Length; i++)
            {
                Console.Write(" ");
            }
            Console.WriteLine(" {0}\t{1}\t    {2}\t{3}", domid, (vm.memoryDynamicMax / 1024 / 1024), vm.VCPUsAtStartup, state);
        }

        return 0;
    }

    public static int dmesg(XenAPI x, bool clear_buffer)
    {
        XenAPI.Host host = new XenAPI.Host(x);
        string message;

        message = host.GetDmesg(clear_buffer);
        if (message == null)
        {
            return -1;
        }
        else if (message == "")
        {
            // nothing
            return 0;
        }

        Console.WriteLine(message);
        return 0;
    }

    public static int log(XenAPI x)
    {
        XenAPI.Host host = new XenAPI.Host(x);
        string message;

        message = host.GetLog();
        if (message == null)
        {
            return -1;
        }
        else if (message == "")
        {
            // nothing
            return 0;
        }

        Console.WriteLine(message);
        return 0;
    }

    public static int destroy(XenAPI x, string target)
    {
        bool isSuccessScanVM = x.ScanVM();
        if (!isSuccessScanVM)
        {
            Console.WriteLine("ERROR: Could not get VMs information.");
            return 1;
        }

        if (_IsDom0(target))
        {
            Console.WriteLine("ERROR: Cannot destroy privileged domain 0");
            return 1;
        }

        ArrayList VMs;
        VMs = x.GetAllVMs();
        foreach (XenAPI.VM vm in VMs)
        {
            if (_IsTarget(vm, target))
            {
                vm.HardShutdown();
                return 0;
            }
        }

        Console.WriteLine("ERROR: Domain '{0}' does not exist.", target);
        return 1;
    }

    public static int shutdown(XenAPI x, string target)
    {
        bool isSuccessScanVM = x.ScanVM();
        if (!isSuccessScanVM)
        {
            Console.WriteLine("ERROR: Could not get VMs information.");
            return 1;
        }

        if (_IsDom0(target))
        {
            Console.WriteLine("ERROR: Domain 0 cannot be shutdown");
            return 1;
        }

        ArrayList VMs;
        VMs = x.GetAllVMs();
        foreach (XenAPI.VM vm in VMs)
        {
            if (_IsTarget(vm, target))
            {
                vm.CleanShutdown();
                return 0;
            }
        }

        Console.WriteLine("ERROR: Domain '{0}' does not exist.", target);
        return 1;
    }

    public static int start(XenAPI x, string target, bool paused)
    {
        bool isSuccessScanVM = x.ScanVM();
        if (!isSuccessScanVM)
        {
            Console.WriteLine("ERROR: Could not get VMs information.");
            return 1;
        }

        if (_IsDom0(target))
        {
            Console.WriteLine("ERROR: Domain is already running");
            return 1;
        }

        ArrayList VMs;
        VMs = x.GetAllVMs();
        foreach (XenAPI.VM vm in VMs)
        {
            if (_IsTarget(vm, target))
            {
                vm.Start(paused);
                return 0;
            }
        }

        Console.WriteLine("ERROR: Domain '{0}' does not exist.", target);
        return 1;
    }

    public static int reboot(XenAPI x, string target)
    {
        bool isSuccessScanVM = x.ScanVM();
        if (!isSuccessScanVM)
        {
            Console.WriteLine("ERROR: Could not get VMs information.");
            return 1;
        }
        
        if (_IsDom0(target))
        {
            Console.WriteLine("ERROR: Domain 0 cannot be shutdown");
            return 1;
        }

        ArrayList VMs;
        VMs = x.GetAllVMs();
        foreach (XenAPI.VM vm in VMs)
        {
            if (_IsTarget(vm, target))
            {
                vm.CleanReboot();
                return 0;
            }
        }

        Console.WriteLine("ERROR: Domain '{0}' does not exist.", target);
        return 1;
    }

    public static int suspend(XenAPI x, string target)
    {
        bool isSuccessScanVM = x.ScanVM();
        if (!isSuccessScanVM)
        {
            Console.WriteLine("ERROR: Could not get VMs information.");
            return 1;
        }

        if (_IsDom0(target))
        {
            Console.WriteLine("ERROR: Cannot save privileged domain 0");
            return 1;
        }

        ArrayList VMs;
        VMs = x.GetAllVMs();
        foreach (XenAPI.VM vm in VMs)
        {
            if (_IsTarget(vm, target))
            {
                vm.Suspend();
                return 0;
            }
        }

        Console.WriteLine("ERROR: Domain '{0}' does not exist.", target);
        return 1;
    }

    public static int resume(XenAPI x, string target, bool start_paused)
    {
        bool isSuccessScanVM = x.ScanVM();
        if (!isSuccessScanVM)
        {
            Console.WriteLine("ERROR: Could not get VMs information.");
            return 1;
        }
        
        if (_IsDom0(target))
        {
            Console.WriteLine("ERROR: Error occurred when resuming: Cannot save privileged domain 0");
            return 1;
        }

        ArrayList VMs;
        VMs = x.GetAllVMs();
        foreach (XenAPI.VM vm in VMs)
        {
            if (_IsTarget(vm, target))
            {
                vm.Resume(start_paused);
                return 0;
            }
        }

        Console.WriteLine("ERROR: Domain '{0}' does not exist.", target);
        return 1;
    }

    public static int pause(XenAPI x, string target)
    {
        bool isSuccessScanVM = x.ScanVM();
        if (!isSuccessScanVM)
        {
            Console.WriteLine("ERROR: Could not get VMs information.");
            return 1;
        }

        if (_IsDom0(target))
        {
            Console.WriteLine("ERROR: Cannot pause privileged domain 0");
            return 1;
        }

        ArrayList VMs;
        VMs = x.GetAllVMs();
        foreach (XenAPI.VM vm in VMs)
        {
            if (_IsTarget(vm, target))
            {
                vm.Pause();
                return 0;
            }
        }

        Console.WriteLine("ERROR: Domain '{0}' does not exist.", target);
        return 1;
    }

    public static int unpause(XenAPI x, string target)
    {
        bool isSuccessScanVM = x.ScanVM();
        if (!isSuccessScanVM)
        {
            Console.WriteLine("ERROR: Could not get VMs information.");
            return 1;
        }

        if (_IsDom0(target))
        {
            Console.WriteLine("ERROR: Cannot unpause privileged domain 0");
            return 1;
        }

        ArrayList VMs;
        VMs = x.GetAllVMs();
        foreach (XenAPI.VM vm in VMs)
        {
            if (_IsTarget(vm, target))
            {
                vm.Unpause();
                return 0;
            }
        }

        Console.WriteLine("ERROR: Domain '{0}' does not exist.", target);
        return 1;
    }

    public static int migrate(XenAPI x, string target, string dest, int port, bool live)
    {
        bool isSuccessScanVM = x.ScanVM();
        if (!isSuccessScanVM)
        {
            Console.WriteLine("ERROR: Could not get VMs information.");
            return 1;
        }

        if (_IsDom0(target))
        {
            Console.WriteLine("ERROR: Cannot unpause privileged domain 0");
            return 1;
        }

        ArrayList VMs;
        VMs = x.GetAllVMs();
        foreach (XenAPI.VM vm in VMs)
        {
            if (_IsTarget(vm, target))
            {
                vm.Migrate(dest, port, live);
                return 0;
            }
        }

        Console.WriteLine("ERROR: Domain '{0}' does not exist.", target);
        return 1;
    }
    
    public static void help()
    {
        Console.WriteLine("Usage: xm <subcommand> [args]");
        Console.WriteLine("");
        Console.WriteLine("Control, list, and manipulate Xen guest instances.");
        Console.WriteLine("");
        Console.WriteLine("xm full list of subcommands:");
        Console.WriteLine("");
        Console.WriteLine(" connect\t\tConnect a xen daemon.");
        Console.WriteLine(" disconnect\t\tDisconnect a xen daemon.");
        Console.WriteLine(" domid\t\t\tList information about all/some domains.");
        Console.WriteLine(" domname\t\tShutdown a domain.");
        Console.WriteLine(" list\t\t\tList information about all/some domains.");
        Console.WriteLine(" destroy\t\tTerminate a domain immediately.");
        Console.WriteLine(" resume\t\t\tResume a Xend managed domain");
        Console.WriteLine(" shutdown\t\tShutdown a domain.");
        Console.WriteLine(" start\t\t\tStart a Xend managed domain");
        Console.WriteLine(" suspend\t\tSuspend a Xend managed domain");
        Console.WriteLine(" pause\t\t\tPause execution of a domain.");
        Console.WriteLine(" unpause\t\tUnpause a paused domain.");
        Console.WriteLine(" migrate\t\tMigrate a domain to another machine.");
        Console.WriteLine(" dmesg\t\t\tRead and/or clear Xend's message buffer.");
        Console.WriteLine(" log\t\t\tPrint Xend log");
        Console.WriteLine("");
        Console.WriteLine("<Domain> can either be the Domain Name or Id.");
        Console.WriteLine("For more help on 'xm' see the xm(1) man page.");
        Console.WriteLine("For more help on 'xm create' see the xmdomain.cfg(5)  man page.");
    }
}